@@ -3,21 +3,23 @@
 The main entry point should be through the Client. The other
 classes provide more modular functionality.
 """
-from rime_sdk.client import Client, RIMEClient
+from rime_sdk.client import Client, ImageType, RIMEClient
 from rime_sdk.firewall import Firewall
 from rime_sdk.image_builder import RIMEImageBuilder
-from rime_sdk.job import Job
+from rime_sdk.job import ContinuousTestJob, Job
 from rime_sdk.project import Project
 from rime_sdk.protos.image_registry.image_registry_pb2 import ManagedImage
 from rime_sdk.protos.model_testing.model_testing_pb2 import CustomImage
-from rime_sdk.test_run import TestRun
+from rime_sdk.test_run import ContinuousTestRun, TestRun
 from rime_sdk.tests import TestBatch
 
 __all__ = [
     "Client",
     "Project",
     "Job",
+    "ContinuousTestJob",
     "TestRun",
+    "ContinuousTestRun",
     "TestBatch",
     "Firewall",
     "RIMEImageBuilder",
