@@ -1,19 +1,34 @@
 """Library for connecting to RIME backend services."""
 
-from typing import Any, Callable, Generic, Optional, TypeVar
+import logging
+from typing import Any, Callable, Dict, Generic, Optional, TypeVar
 
 import grpc
+import importlib_metadata
 
+from rime_sdk.internal.client_interceptor import (
+    AddMetadataUnaryStreamClientInterceptor,
+    AddMetadataUnaryUnaryClientInterceptor,
+)
+from rime_sdk.internal.grpc_error_handler import GRPCErrorHandler
+from rime_sdk.protos.data_collector.data_collector_pb2_grpc import DataCollectorStub
+from rime_sdk.protos.feature_flag.feature_flag_pb2_grpc import FeatureFlagStub
 from rime_sdk.protos.file_upload.file_upload_pb2_grpc import FileUploadStub
 from rime_sdk.protos.firewall.firewall_pb2_grpc import FirewallServiceStub
 from rime_sdk.protos.image_registry.image_registry_pb2_grpc import ImageRegistryStub
+from rime_sdk.protos.jobs.jobs_pb2_grpc import JobReaderStub
 from rime_sdk.protos.model_testing.model_testing_pb2_grpc import ModelTestingStub
+from rime_sdk.protos.notification.notification_pb2_grpc import NotificationSettingStub
 from rime_sdk.protos.project.project_pb2_grpc import ProjectManagerStub
+from rime_sdk.protos.rime_info.rime_info_pb2 import GetRIMEInfoRequest
+from rime_sdk.protos.rime_info.rime_info_pb2_grpc import RIMEInfoStub
 from rime_sdk.protos.test_run_results.test_run_results_pb2_grpc import ResultsReaderStub
 from rime_sdk.protos.test_run_tracker.test_run_tracker_pb2_grpc import (
     TestRunTrackerStub,
 )
 
+logger = logging.getLogger()
+
 # Generic type representing a client stub for a gRPC server.
 C = TypeVar("C")
 
@@ -46,6 +61,7 @@ class RIMEConnection(Generic[C]):
         api_key: str,
         channel_timeout: float = 5.0,
         disable_tls: bool = False,
+        ssl_config: Optional[Dict[str, str]] = None,
     ) -> None:
         """Create a new connection for a RIME backend.
 
@@ -61,6 +77,9 @@ class RIMEConnection(Generic[C]):
                 The timeout in seconds for waiting for the given channel.
             disable_tls: bool
                 Whether to disable tls when connecting to the backend.
+            ssl_config: dict(str)
+                SSL config to be passed to grpc.ssl_channel_credentials. Documentation
+                can be found here: https://grpc.github.io/grpc/python/_modules/grpc.html#ssl_channel_credentials
         """
         self._create_backend_fn = create_backend_fn
         self._api_key = api_key
@@ -68,6 +87,7 @@ class RIMEConnection(Generic[C]):
         self._channel_timeout = channel_timeout
         self._channel: Optional[grpc.Channel] = None
         self._disable_tls = disable_tls
+        self._ssl_config = ssl_config if ssl_config is not None else {}
 
     def __enter__(self) -> C:
         """Acquires the channel created in the with-context."""
@@ -103,6 +123,7 @@ class RIMEConnection(Generic[C]):
             ValueError
                 If a connection cannot be made to a backend service within `timeout`.
         """
+
         try:
             # create credentials
             if self._disable_tls:
@@ -113,25 +134,37 @@ class RIMEConnection(Generic[C]):
                     grpc.metadata_call_credentials(GrpcApiKeyAuth(self._api_key)),
                 )
                 channel = grpc.secure_channel(addr, credentials)
+
+            channel = grpc.intercept_channel(
+                channel,
+                AddMetadataUnaryUnaryClientInterceptor(),
+                AddMetadataUnaryStreamClientInterceptor(),
+            )
             grpc.channel_ready_future(channel).result(timeout=timeout)
             return channel
         except grpc.FutureTimeoutError:
-            raise ValueError(f"Could not connect to server at address `{addr}`")
+            raise ValueError(
+                f"Could not connect to server at address `{addr}`. "
+                "Please confirm the URL is correct and check your network connection."
+            ) from None
 
     def _get_ssl_channel_credentials(self) -> grpc.ChannelCredentials:
         """Fetch channel credentials for an SSL channel."""
-        return grpc.ssl_channel_credentials()
+        return grpc.ssl_channel_credentials(**self._ssl_config)
 
 
 class RIMEBackend:
     """An abstraction for connecting to RIME's backend services."""
 
+    GRPCErrorHandler = GRPCErrorHandler
+
     def __init__(
         self,
         domain: str,
         api_key: str = "",
         channel_timeout: float = 5.0,
         disable_tls: bool = False,
+        ssl_config: Optional[Dict[str, str]] = None,
     ):
         """Create a new RIME backend.
 
@@ -145,15 +178,42 @@ class RIMEBackend:
                 when opening connections to gRPC servers.
             disable_tls: bool
                 Whether to disable tls when connecting to the backend.
+            ssl_config: dict(str)
+                SSL config to be passed to grpc.ssl_channel_credentials. Documentation
+                can be found here: https://grpc.github.io/grpc/python/_modules/grpc.html#ssl_channel_credentials
         """
         self._channel_timeout = channel_timeout
         self._api_key = api_key
         self._disable_tls = disable_tls
         domain_split = domain.split(".", 1)
+        self._ssl_config = ssl_config if ssl_config is not None else {}
         if domain_split[0][-4:] == "rime":
             base_domain = domain_split[1]
             domain = "rime-backend." + base_domain
+        if domain.endswith("/"):
+            domain = domain[:-1]
         self._backend_addr = self._get_backend_addr(domain, disable_tls)
+        # Make sure this is last as it is dependent on connections being configured.
+        self.customer_name = self._check_version()
+
+    def _check_version(self) -> str:
+        """Check current RIME version and return client name."""
+        with self.get_rime_info_stub() as rime_info_client:
+            with self.GRPCErrorHandler():
+                rime_info_response = rime_info_client.GetRIMEInfo(GetRIMEInfoRequest())
+
+        server_version = rime_info_response.cluster_info_version
+        client_version = importlib_metadata.version("rime_sdk")
+        if client_version != server_version:
+            logger.warning(
+                "Python SDK package and server are on different versions. "
+                f"The Python SDK package is on version {client_version}, "
+                f"while the server is on version {server_version}. "
+                f"In order to make them be on the same version, please "
+                f"install the correct version of the Python SDK package with "
+                f"`pip install rime_sdk=={server_version}`"
+            )
+        return rime_info_response.customer_name
 
     def _get_backend_addr(self, domain: str, disable_tls: bool = False) -> str:
         """Construct an address to the all backend services from `domain`.
@@ -178,6 +238,7 @@ class RIMEBackend:
             self._api_key,
             channel_timeout=self._channel_timeout,
             disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
         )
 
     def get_image_registry_stub(self) -> RIMEConnection[ImageRegistryStub]:
@@ -188,6 +249,7 @@ class RIMEBackend:
             self._api_key,
             channel_timeout=self._channel_timeout,
             disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
         )
 
     def get_model_testing_stub(self) -> RIMEConnection[ModelTestingStub]:
@@ -198,6 +260,7 @@ class RIMEBackend:
             self._api_key,
             channel_timeout=self._channel_timeout,
             disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
         )
 
     def get_test_run_tracker_stub(self) -> RIMEConnection[TestRunTrackerStub]:
@@ -208,6 +271,7 @@ class RIMEBackend:
             self._api_key,
             channel_timeout=self._channel_timeout,
             disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
         )
 
     def get_test_run_results_stub(self) -> RIMEConnection[ResultsReaderStub]:
@@ -218,6 +282,7 @@ class RIMEBackend:
             self._api_key,
             channel_timeout=self._channel_timeout,
             disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
         )
 
     def get_firewall_stub(self) -> RIMEConnection[FirewallServiceStub]:
@@ -228,6 +293,18 @@ class RIMEBackend:
             self._api_key,
             channel_timeout=self._channel_timeout,
             disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
+        )
+
+    def get_data_collector_stub(self) -> RIMEConnection[DataCollectorStub]:
+        """Return a Data Collector client."""
+        return RIMEConnection[DataCollectorStub](
+            DataCollectorStub,
+            self._backend_addr,
+            self._api_key,
+            channel_timeout=self._channel_timeout,
+            disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
         )
 
     def get_project_manager_stub(self) -> RIMEConnection[ProjectManagerStub]:
@@ -238,4 +315,49 @@ class RIMEBackend:
             self._api_key,
             channel_timeout=self._channel_timeout,
             disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
+        )
+
+    def get_feature_flag_stub(self) -> RIMEConnection[FeatureFlagStub]:
+        """Return a feature flag client."""
+        return RIMEConnection[FeatureFlagStub](
+            FeatureFlagStub,
+            self._backend_addr,
+            self._api_key,
+            channel_timeout=self._channel_timeout,
+            disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
+        )
+
+    def get_rime_info_stub(self) -> RIMEConnection[RIMEInfoStub]:
+        """Return a rime info client."""
+        return RIMEConnection[RIMEInfoStub](
+            RIMEInfoStub,
+            self._backend_addr,
+            self._api_key,
+            channel_timeout=self._channel_timeout,
+            disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
+        )
+
+    def get_job_reader_stub(self) -> RIMEConnection[JobReaderStub]:
+        """Return a job reader client."""
+        return RIMEConnection[JobReaderStub](
+            JobReaderStub,
+            self._backend_addr,
+            self._api_key,
+            channel_timeout=self._channel_timeout,
+            disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
+        )
+
+    def get_notification_settings_stub(self) -> RIMEConnection[NotificationSettingStub]:
+        """Return a Notification setting client."""
+        return RIMEConnection[NotificationSettingStub](
+            NotificationSettingStub,
+            self._backend_addr,
+            self._api_key,
+            channel_timeout=self._channel_timeout,
+            disable_tls=self._disable_tls,
+            ssl_config=self._ssl_config,
         )
